# standard library
from typing import List, Union, Tuple


# dependent packages
import numpy as np
from .physics import c, e, h,johnson_nyquist_psd 


# type aliases
ArrayLike = Union[np.ndarray, List[float], List[int], float, int]


# constants
Delta_Al = 188.0 * 10**-6 * e  # gap energy of Al
eta_pb = 0.4  # Pair breaking efficiency
eta_Al_ohmic_850 = 0.9975  # Ohmic loss of an Al surface at 850 GHz.
# Shitov+, ISSTT2008. https://www.nrao.edu/meetings/isstt/papers/2008/2008263266.pdf


# main functions
def D2HPBW(F: ArrayLike) -> ArrayLike:
    """Get half-power beam width of DESHIMA 2.0 at given frequency (frequencies).

    Parameters
    ----------
    F
        Frequency. Units: Hz.

    Returns
    -------
    hpbw
        Half-power beam width. Units: radian.

    """
    return 29.0 * 240.0 / (F / 1e9) * np.pi / 180.0 / 60.0 / 60.0


def eta_mb_ruze(F: ArrayLike, LFlimit: float, sigma: float) -> ArrayLike:
    """Get main-beam efficiency by Ruze's equation.

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    LFlimit
        Main-beam efficiency at 0 Hz.
    sigma
        Surface error. Units: m.

    Returns
    -------
    eta_mb
        Main-beam efficiency. Units: None.

    """
    return LFlimit * np.exp(-((4.0 * np.pi * sigma * F / c) ** 2.0))


def photon_NEP_kid(F: ArrayLike, 
                   psd_KID: ArrayLike) -> ArrayLike:
    """NEP of the KID, with respect to the absorbed power.

    Parameters
    -----------
    F
        Frequency of the signal responsible for loading. Units: Hz.
    psd_kid
        Power spectral density entering MKID the KID. Units: W / Hz.
    filterbank
        Detection bandwidth, with respect to the power that sets the loading. Units: Hz.

    Returns
    -------
    NEP_kid
        Noise-equivalent power of the KID.

    Notes
    -----
    Pkid/(W_F * h * F) gives the occupation number.

    """
    # photon_term = 2 * Pkid * (h*F + Pkid/W_F)
    poisson_term = 2 * psd_KID * h * F
    bunching_term = 2 * psd_KID**2
    r_term = 4 * Delta_Al * psd_KID / eta_pb
    return poisson_term + bunching_term + r_term


def window_trans(
    F: ArrayLike,
    thickness: ArrayLike,
    tandelta: float,
    neffHDPE: float,
    window_AR: bool,
    T_parasitic_refl: float,
    T_parasitic_refr: float
) -> Tuple[ArrayLike, ArrayLike]:
    """Calculates the window transmission.

    Parameters
    ----------
    F
        Frequency. Units: Hz.
    thickness
        Thickness of the HDPE window. Units: m.
    tandelta
        Loss tangent of window/lens dielectric.
    neffHDPE
        Refractive index of HDPE. Set to 1 to remove reflections. Units : None.
    window_AR
        Whether the window is supposed to be coated by Ar (True) or not (False).
    T_parasitic_refl
        Temperature of parasitic source seen in reflection, w.r.t. instrument.
    T_parasitic_refr
        Temperature of parasitic source seen in refraction..


    Returns
    -------
    List containing list of arrays of efficiencies as first element, and list of arrays of psd's seen by each stage.
    """

    eta = []
    psd = []
    
    HDPErefl = ((1 - neffHDPE) / (1 + neffHDPE)) ** 2 * np.ones(F.size)
    psd_refl = johnson_nyquist_psd(F, T_parasitic_refl)
    psd_refr = johnson_nyquist_psd(F, T_parasitic_refr)

    if window_AR == False:
        eta.append(1 - HDPErefl)
        psd.append(psd_refl)

    eta_HDPE = np.exp(
        -thickness
        * 2
        * np.pi
        * neffHDPE
        * (tandelta * F / c + (tandelta * F / c) ** 2)
    )

    eta.append(eta_HDPE)
    psd.append(psd_refr)

    if window_AR == False:
        eta.append(1 - HDPErefl)
        psd.append(psd_refl)

    return eta, psd

def average_over_filterbank(array_to_average: ArrayLike,
                            filterbank: ArrayLike) -> ArrayLike:
    """Averages a quantity defined on F_sky over as filterbank.

    Parameters
    ----------
    array_to_average
        Array defined over F_sky to average
    filterbank
        2D array containing filterbank

    Returns
    -------
    Averaged array.
    """
    sh_f = filterbank.shape
    assert array_to_average.size == sh_f[0]

    array_tiled = np.tile(array_to_average, (sh_f[-1], 1)) * filterbank.T
    return np.nansum(array_tiled, axis=1) / np.nansum(filterbank, axis=0)



import numpy as np

from typing import List, Union, Dict, Tuple

from .instruments import eta_Al_ohmic_850, photon_NEP_kid, window_trans
from .physics import johnson_nyquist_psd

def eta_Al_ohmic(F_sky: np.ndarray) -> np.ndarray: 
    """Calculate Ohmic losses for aluminium over array of sky frequencies.
    
    Parameters
    ----------
    F_sky
        Numpy array containing sky frequencies. Units: GHz
    
    Returns
    ----------
    Array with eta values for Ohmic losses.
    """

    return 1.0 - (1.0 - eta_Al_ohmic_850) * np.sqrt(F_sky / 850)

def sizer(eta: Union[np.ndarray, float], 
           F_sky: np.ndarray, 
           F_eta: np.ndarray = None
) -> np.ndarray:
    """Resize efficiency term to new size.

    Used to vectorize or interpolate on efficiency terms.
    If efficiency is a scalar, an array is returned with the same size as F_sky.
    If efficiency is an array with different size then F_sky, an array containing frequencies at which eta is evaluated should also be passed.
    A 1D interpolation on F_sky is then performed to evaluate eta on F_sky.
    If efficiency is array with same size as F_sky, it is returned as-is. 
    Responisibility to verify if the efficiencies are evaluated on the same frequencies as present in F_sky is placed on the user.

    Parameters
    ----------
    eta
        Efficiency term.
    F_sky
        Numpy array containing sky frequencies. Units: GHz
    F_eta
        Numpy array containing frequencies at which eta is evaluated.
        Should only be passed when 1D interpolation is required and defaults to None.
    
    Returns
    ----------
    Array with eta values, depending on input (see above).
    """

    if not hasattr(eta, "__len__"):
        return eta * np.ones(F_sky.size)

    elif F_eta is not None:
        idx_sorted = np.argsort(F_eta)
        return np.interp(F_sky, 
                         F_eta[idx_sorted], 
                         eta[idx_sorted])

    else:
        return eta

def get_cascade(cascade_list: List[Dict[any, any]],
                F_sky: np.ndarray,
                ) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate a cascade list, consisting of efficiency and psd per stage.

    Parameters
    ----------
    cascade_list
        List containing, per element, the efficiency and coupling temperature of each stage in the cascade.
        For reflective stages, the dictionary should contain either:
            - A single eta and temperature
            - A tuple with efficiencies and frequencies at which these are defines, and a temperature
        
        For refractive stages, the dictionary should contain:
            - thickness of dielectric in meters, loss tangent, effective refractive index, whether to use AR coating, temperature seen in reflection coming from the ISS, and temperature seen in refraction.

    F_sky
        Array with sky frequencies. Units: GHz.
    
    Returns
    ----------
    List with list of arrays containing efficiencies as first element, and list containing arrays of psd as second element. 
    """


    group_list = []
    
    for cascade in cascade_list:
        group_list.append(cascade.get("group"))
    
    group_list_red_uniq = list(dict.fromkeys([x for x in group_list if x is not None]))

    idx_group_list = [i if label == None else label for i, label in enumerate(group_list)]
    cascade_type_list = np.array([0 if x.get("eta_coup") is not None else 1 for x in cascade_list])
    idx_group_list_uniq = list(dict.fromkeys(idx_group_list))

    to_delete = []
    for group_label in group_list_red_uniq:
        index_list = [i for i, label in enumerate(idx_group_list) if label == group_label]
        idx_group_list_uniq[idx_group_list_uniq.index(group_label)] = index_list

        to_delete.extend(index_list[:-1])

    cascade_type_list_uniq = np.delete(cascade_type_list, to_delete)
    idx_group_list_uniq = [[x] if hasattr(x, "__len__") == False else x for x in idx_group_list_uniq] 

    # Now get eta for groups

    group_running = None
    all_eta = []
    all_psd = []
    use_for_eta_inst = []
    eta_inst_flag = 0

    for casc_t, idx_group in zip(cascade_type_list_uniq, idx_group_list_uniq):
        # This cascade group is reflective
        if casc_t == 0:
            eta_grouped = np.ones(F_sky.size)

            if (T_casc := cascade_list[idx_group[0]].get("T_parasitic")) == "atmosphere":
                all_psd.append(T_casc) # Group couples to atmosphere: calculate psd in rad trans loop.
            else:
                all_psd.append(johnson_nyquist_psd(F_sky*1e9, T_casc)) # Calculate psd for T_parasitic

            for idx_g in idx_group:
                eta_interp_flag = False
                
                if isinstance(eta := cascade_list[idx_g].get("eta_coup"), tuple): # eta_coup is a tuple; array of eta and frequencies
                    assert(len(eta) == 2)
                    
                    eta, F_eta = eta

                    assert(eta.size == F_eta.size)
                    
                    eta_interp_flag = True

                elif eta == "Ohmic-Al": # generate vector with eta of Aluminium
                    eta = eta_Al_ohmic(F_sky) 

                elif eta_interp_flag:
                    eta_grouped *= sizer(eta, F_sky, F_eta)
                else:
                    print(type(eta))
                    print(eta)
                    eta_grouped *= sizer(eta, F_sky)
            
            all_eta.append(eta_grouped)
            
            use_for_eta_inst.append(eta_inst_flag)
        
        if casc_t == 1:
            for idx_g in idx_group:
                if (casc := cascade_list[idx_g]).get("cryo_window_flag"):
                    eta_inst_flag = 1
                etas, psds = window_trans(F_sky * 1e9,
                                          casc.get("thickness"), 
                                          casc.get("tandelta"), 
                                          casc.get("neff"), 
                                          casc.get("window_AR"), 
                                          casc.get("T_parasitic_refl"), 
                                          casc.get("T_parasitic_refr")) 

                all_eta.extend(etas)
                all_psd.extend(psds)
        
                use_for_eta_inst.extend([eta_inst_flag for _ in range(len(etas))])
    
    return all_eta, all_psd, use_for_eta_inst
